@@ -1,6 +1,6 @@
 from abc import ABC, abstractmethod
 from dataclasses import dataclass
-from typing import List, Optional
+from typing import List, Optional, Union
 
 from netqasm.lang import encoding
 from netqasm.lang.operand import (
@@ -43,7 +43,7 @@ class NetQASMInstruction(ABC):
 
     @classmethod
     @abstractmethod
-    def from_operands(cls, operands: List[Operand]) -> "NetQASMInstruction":
+    def from_operands(cls, operands: List[Union[Operand, int]]) -> "NetQASMInstruction":
         pass
 
     def writes_to(self) -> List[Register]:
@@ -96,7 +96,7 @@ class NoOperandInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 0
         return cls(id=cls.id)
 
@@ -128,7 +128,7 @@ class RegInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 1
         reg = operands[0]
         assert isinstance(reg, Register)
@@ -166,7 +166,7 @@ class RegRegInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 2
         reg0, reg1 = operands
         assert isinstance(reg0, Register)
@@ -207,13 +207,26 @@ class RegImmImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 3
-        reg, imm0, imm1 = operands
+        reg = operands[0]
         assert isinstance(reg, Register)
-        assert isinstance(imm0, int)
-        assert isinstance(imm1, int)
-        return cls(reg=reg, imm0=Immediate(value=imm0), imm1=Immediate(value=imm1))
+
+        imm0 = (
+            Immediate(value=operands[1])
+            if isinstance(operands[1], int)
+            else operands[1]
+        )
+        assert isinstance(imm0, Immediate)
+
+        imm1 = (
+            Immediate(value=operands[2])
+            if isinstance(operands[2], int)
+            else operands[2]
+        )
+        assert isinstance(imm1, Immediate)
+
+        return cls(reg=reg, imm0=imm0, imm1=imm1)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.reg)} {str(self.imm0)} {str(self.imm1)}"
@@ -255,21 +268,91 @@ class RegRegImmImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 4
         reg0, reg1, imm0, imm1 = operands
         assert isinstance(reg0, Register)
         assert isinstance(reg1, Register)
-        assert isinstance(imm0, int)
-        assert isinstance(imm1, int)
-        return cls(
-            reg0=reg0, reg1=reg1, imm0=Immediate(value=imm0), imm1=Immediate(value=imm1)
-        )
+        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
+        if isinstance(imm0, int):
+            imm0 = Immediate(value=imm0)
+        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
+        if isinstance(imm1, int):
+            imm1 = Immediate(value=imm1)
+        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm0)} {str(self.imm1)}"
 
 
+@dataclass
+class RegRegImm4Instruction(NetQASMInstruction):
+    """
+    An instruction with 2 Register operands followed by 4 Immediate operands.
+    """
+
+    reg0: Register = None  # type: ignore
+    reg1: Register = None  # type: ignore
+    imm0: Immediate = None  # type: ignore
+    imm1: Immediate = None  # type: ignore
+    imm2: Immediate = None  # type: ignore
+    imm3: Immediate = None  # type: ignore
+
+    @property
+    def operands(self) -> List[Operand]:
+        return [self.reg0, self.reg1, self.imm0, self.imm1, self.imm2, self.imm3]
+
+    @classmethod
+    def deserialize_from(cls, raw: bytes):
+        c_struct = encoding.RegRegImm4Command.from_buffer_copy(raw)
+        assert c_struct.id == cls.id
+        reg0 = Register.from_raw(c_struct.reg0)
+        reg1 = Register.from_raw(c_struct.reg1)
+        imm0 = Immediate(value=c_struct.imm0)
+        imm1 = Immediate(value=c_struct.imm1)
+        imm2 = Immediate(value=c_struct.imm2)
+        imm3 = Immediate(value=c_struct.imm3)
+        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1, imm2=imm2, imm3=imm3)
+
+    def serialize(self) -> bytes:
+        c_struct = encoding.RegRegImm4Command(
+            id=self.id,
+            reg0=self.reg0.cstruct,
+            reg1=self.reg1.cstruct,
+            imm0=self.imm0.value,
+            imm1=self.imm1.value,
+            imm2=self.imm2.value,
+            imm3=self.imm3.value,
+        )
+        return bytes(c_struct)
+
+    @classmethod
+    def from_operands(cls, operands: List[Union[Operand, int]]):
+        assert len(operands) == 6
+        reg0, reg1, imm0, imm1, imm2, imm3 = operands
+        assert isinstance(reg0, Register)
+        assert isinstance(reg1, Register)
+        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
+        if isinstance(imm0, int):
+            imm0 = Immediate(value=imm0)
+        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
+        if isinstance(imm1, int):
+            imm1 = Immediate(value=imm1)
+        assert isinstance(imm2, int) or isinstance(imm2, Immediate)
+        if isinstance(imm2, int):
+            imm2 = Immediate(value=imm2)
+        assert isinstance(imm3, int) or isinstance(imm3, Immediate)
+        if isinstance(imm3, int):
+            imm3 = Immediate(value=imm3)
+        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1, imm2=imm2, imm3=imm3)
+
+    def _pretty_print(self):
+        return (
+            f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm0)} "
+            f"{str(self.imm1)} {str(self.imm2)} {str(self.imm3)}"
+        )
+
+
 @dataclass
 class RegRegRegInstruction(NetQASMInstruction):
     """
@@ -303,7 +386,7 @@ class RegRegRegInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 3
         reg0, reg1, reg2 = operands
         assert isinstance(reg0, Register)
@@ -351,7 +434,7 @@ class RegRegRegRegInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 4
         reg0, reg1, reg2, reg3 = operands
         assert isinstance(reg0, Register)
@@ -393,11 +476,14 @@ class ImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 1
         imm = operands[0]
-        assert isinstance(imm, int)
-        return cls(imm=Immediate(value=imm))
+        assert isinstance(imm, int) or isinstance(imm, Immediate)
+        if isinstance(imm, int):
+            imm = Immediate(value=imm)
+
+        return cls(imm=imm)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.imm)}"
@@ -431,13 +517,17 @@ class ImmImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 2
         imm0 = operands[0]
         imm1 = operands[1]
-        assert isinstance(imm0, int)
-        assert isinstance(imm1, int)
-        return cls(imm0=Immediate(value=imm0), imm1=Immediate(value=imm1))
+        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
+        if isinstance(imm0, int):
+            imm0 = Immediate(value=imm0)
+        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
+        if isinstance(imm1, int):
+            imm1 = Immediate(value=imm1)
+        return cls(imm0=imm0, imm1=imm1)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.imm0)} {str(self.imm1)}"
@@ -476,13 +566,15 @@ class RegRegImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 3
         reg0, reg1, imm = operands
         assert isinstance(reg0, Register)
         assert isinstance(reg1, Register)
-        assert isinstance(imm, int)
-        return cls(reg0=reg0, reg1=reg1, imm=Immediate(value=imm))
+        assert isinstance(imm, int) or isinstance(imm, Immediate)
+        if isinstance(imm, int):
+            imm = Immediate(value=imm)
+        return cls(reg0=reg0, reg1=reg1, imm=imm)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm)}"
@@ -516,12 +608,14 @@ class RegImmInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 2
         reg, imm = operands
         assert isinstance(reg, Register)
-        assert isinstance(imm, int)
-        return cls(reg=reg, imm=Immediate(value=imm))
+        assert isinstance(imm, int) or isinstance(imm, Immediate)
+        if isinstance(imm, int):
+            imm = Immediate(value=imm)
+        return cls(reg=reg, imm=imm)
 
     def _pretty_print(self):
         return f"{self.mnemonic} {str(self.reg)} {str(self.imm)}"
@@ -555,7 +649,7 @@ class RegEntryInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 2
         reg, entry = operands
         assert isinstance(reg, Register)
@@ -594,7 +688,7 @@ class RegAddrInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 2
         reg, addr = operands
         assert isinstance(reg, Register)
@@ -629,7 +723,7 @@ class ArrayEntryInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 1
         entry = operands[0]
         assert isinstance(entry, ArrayEntry)
@@ -663,7 +757,7 @@ class ArraySliceInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 1
         slice = operands[0]
         assert isinstance(slice, ArraySlice)
@@ -697,7 +791,7 @@ class AddrInstruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 1
         addr = operands[0]
         assert isinstance(addr, Address)
@@ -747,7 +841,7 @@ class Reg5Instruction(NetQASMInstruction):
         return bytes(c_struct)
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         assert len(operands) == 5
         reg0, reg1, reg2, reg3, reg4 = operands
         assert isinstance(reg0, Register)
@@ -784,7 +878,7 @@ class DebugInstruction(NetQASMInstruction):
         return b""
 
     @classmethod
-    def from_operands(cls, operands: List[Operand]):
+    def from_operands(cls, operands: List[Union[Operand, int]]):
         pass
 
     def _pretty_print(self):
