@@ -20,28 +20,27 @@ from typing import (
     Union,
 )
 
-from netqasm import NETQASM_VERSION
 from netqasm.backend.network_stack import OK_FIELDS_K, OK_FIELDS_M
 from netqasm.lang import operand
 from netqasm.lang.encoding import RegisterName
 from netqasm.lang.ir import (
-    Address,
-    ArrayEntry,
-    ArraySlice,
     BranchLabel,
     BreakpointAction,
     BreakpointRole,
     GenericInstr,
     ICmd,
-    Label,
-    PreSubroutine,
-    T_OperandUnion,
+    ProtoSubroutine,
+    T_ProtoOperand,
     flip_branch_instr,
 )
+from netqasm.lang.operand import Address, ArrayEntry, ArraySlice, Label, Template
 from netqasm.lang.parsing.text import assemble_subroutine, parse_register
 from netqasm.lang.subroutine import Subroutine
-from netqasm.qlink_compat import EPRRole, EPRType, LinkLayerOKTypeK
+from netqasm.lang.version import NETQASM_VERSION
+from netqasm.qlink_compat import BellState, EPRRole, EPRType, LinkLayerOKTypeK
 from netqasm.sdk.build_epr import (
+    SER_RESPONSE_KEEP_IDX_BELL_STATE,
+    SER_RESPONSE_KEEP_LEN,
     EntRequestParams,
     EprKeepResult,
     EprMeasureResult,
@@ -59,13 +58,13 @@ from netqasm.sdk.build_types import (
     T_LoopRoutine,
     T_PostRoutine,
 )
-from netqasm.sdk.compiling import NVSubroutineCompiler, SubroutineCompiler
 from netqasm.sdk.config import LogConfig
 from netqasm.sdk.constraint import SdkConstraint, ValueAtMostConstraint
 from netqasm.sdk.futures import Array, Future, RegFuture, T_CValue
 from netqasm.sdk.memmgr import MemoryManager
-from netqasm.sdk.qubit import FutureQubit, Qubit
+from netqasm.sdk.qubit import FutureQubit, Qubit, QubitMeasureBasis
 from netqasm.sdk.toolbox import get_angle_spec_from_float
+from netqasm.sdk.transpile import NVSubroutineTranspiler, SubroutineTranspiler
 from netqasm.typedefs import T_Cmd
 from netqasm.util.log import LineTracker
 
@@ -189,11 +188,11 @@ class SdkLoopUntilContext:
 
 
 class Builder:
-    """Object that transforms Python script code into `PreSubroutine`s.
+    """Object that transforms Python script code into `ProtoSubroutine`s.
 
     A Connection uses a Builder to handle statements in application script code.
     The Builder converts the statements into pseudo-NetQASM instructions that are
-    assembled into a PreSubroutine. When the connectin flushes, the PreSubroutine is
+    assembled into a ProtoSubroutine. When the connectin flushes, the ProtoSubroutine is
     is compiled into a NetQASM subroutine.
     """
 
@@ -203,7 +202,7 @@ class Builder:
         app_id: int,
         hardware_config: Optional[HardwareConfig] = None,
         log_config: LogConfig = None,
-        compiler: Optional[Type[SubroutineCompiler]] = None,
+        compiler: Optional[Type[SubroutineTranspiler]] = None,
         return_arrays: bool = True,
     ):
         """Builder constructor. Typically not used directly by the Host script.
@@ -215,7 +214,7 @@ class Builder:
         :param log_config: logging configuration, typically just passed as-is by the
             connection object
         :param compiler: which compiler class to use for the translation from
-            PreSubroutine to Subroutine
+            ProtoSubroutine to Subroutine
         :param return_arrays: whether to add ret_arr NetQASM instructions at the end of
             each subroutine (for all arrays that are used in the subroutine). May be
             set to False if the quantum node controller does not support returning
@@ -258,11 +257,11 @@ class Builder:
         self._max_qubits: int = self._hardware_config.qubit_count
 
         # What compiler (if any) to be used
-        self._compiler: Optional[Type[SubroutineCompiler]] = compiler
+        self._compiler: Optional[Type[SubroutineTranspiler]] = compiler
 
         # If an NV compiler is specified but not an NV hardware config,
         # make sure an NV config is used after all.
-        if compiler == NVSubroutineCompiler:
+        if compiler == NVSubroutineTranspiler:
             num_qubits = self._hardware_config.qubit_count
             self._hardware_config = NVHardwareConfig(num_qubits)
 
@@ -321,28 +320,23 @@ class Builder:
         self._pending_commands = []
         return commands
 
-    def subrt_pop_pending_subroutine(self) -> Optional[PreSubroutine]:
+    def subrt_pop_pending_subroutine(self) -> Optional[ProtoSubroutine]:
         # Add commands for initialising and returning arrays
         self._build_cmds_allocated_arrays()
         self._build_cmds_return_registers()
         if len(self._pending_commands) > 0:
             commands = self.subrt_pop_all_pending_commands()
-            metadata = self.subrt_get_metadata()
-            return PreSubroutine(**metadata, commands=commands)
+            return ProtoSubroutine(
+                commands=commands, netqasm_version=NETQASM_VERSION, app_id=self.app_id
+            )
         else:
             return None
 
-    def subrt_get_metadata(self) -> Dict:
-        return {
-            "netqasm_version": NETQASM_VERSION,
-            "app_id": self.app_id,
-        }
-
-    def subrt_compile_subroutine(self, pre_subroutine: PreSubroutine) -> Subroutine:
-        """Convert a PreSubroutine into a Subroutine."""
+    def subrt_compile_subroutine(self, pre_subroutine: ProtoSubroutine) -> Subroutine:
+        """Convert a ProtoSubroutine into a Subroutine."""
         subroutine: Subroutine = assemble_subroutine(pre_subroutine)
         if self._compiler is not None:
-            subroutine = self._compiler(subroutine=subroutine).compile()
+            subroutine = self._compiler(subroutine=subroutine).transpile()
         if self._track_lines:
             self._log_subroutine(subroutine=subroutine)
         return subroutine
@@ -364,10 +358,12 @@ class Builder:
         )
 
     def _build_cmds_wait_move_epr_to_mem(
-        self, number: int, ent_results_array: Array
+        self, params: EntRequestParams, ent_results_array: Array
     ) -> None:
 
-        loop_register = self._mem_mgr.get_inactive_register()
+        loop_register = self._mem_mgr.get_inactive_register(activate=True)
+        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
+        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)
 
         def post_loop(conn: BaseNetQASMConnection, loop_reg: RegFuture):
             # Wait for each pair individually
@@ -377,8 +373,15 @@ class Builder:
                 pair=pair,
             )
 
+            if params.expect_phi_plus:
+                # Perform Bell corrections
+                bell_state = self._get_raw_bell_state(
+                    ent_results_array, loop_reg, bell_state_reg
+                )
+                self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
+
             # If it's the last pair, don't move it to a mem qubit
-            with loop_reg.if_ne(number - 1):
+            with loop_reg.if_ne(params.number - 1):
                 reg0 = self._mem_mgr.get_inactive_register(activate=True)
                 reg1 = self._mem_mgr.get_inactive_register(activate=True)
                 assert loop_reg.reg is not None
@@ -387,7 +390,7 @@ class Builder:
                 # It is "number of pairs" - 1 - "current index".
                 sub_cmd = ICmd(
                     instruction=GenericInstr.SUB,
-                    operands=[reg0, number - 1, loop_reg.reg],
+                    operands=[reg0, params.number - 1, loop_reg.reg],
                 )
                 set_0_cmds = ICmd(instruction=GenericInstr.SET, operands=[reg1, 0])
 
@@ -408,20 +411,26 @@ class Builder:
                 self._mem_mgr.remove_active_register(reg0)
                 self._mem_mgr.remove_active_register(reg1)
 
-        self._build_cmds_loop_body(post_loop, stop=number, loop_register=loop_register)
+        self._build_cmds_loop_body(
+            post_loop, stop=params.number, loop_register=loop_register
+        )
+        self._mem_mgr.remove_active_register(loop_register)
+        self._mem_mgr.remove_active_register(qubit_reg)
+        self._mem_mgr.remove_active_register(bell_state_reg)
 
     def _build_cmds_post_epr(
         self,
         qubit_ids: Array,
-        number: int,
+        params: EntRequestParams,
         ent_results_array: Array,
         tp: EPRType,
-        post_routine: T_PostRoutine,
     ) -> None:
 
-        loop_register = self._mem_mgr.get_inactive_register()
+        loop_register = self._mem_mgr.get_inactive_register(activate=True)
+        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
+        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)
 
-        def post_loop(conn: BaseNetQASMConnection, _: RegFuture):
+        def post_loop(conn: BaseNetQASMConnection, loop_reg: RegFuture):
             # Wait for each pair individually
             pair = loop_register
             conn.builder._add_wait_for_ent_info_cmd(
@@ -429,13 +438,26 @@ class Builder:
                 pair=pair,
             )
             assert tp == EPRType.K or tp == EPRType.R
-            q_id = qubit_ids.get_future_index(pair)
+
+            if params.expect_phi_plus:
+                bell_state = self._get_raw_bell_state(
+                    ent_results_array, loop_reg, bell_state_reg
+                )
+                self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
+
+            q_id = qubit_ids.get_future_index(loop_register)
             q = FutureQubit(conn=conn, future_id=q_id)
-            pair_future = RegFuture(self._connection, pair)
-            post_routine(self, q, pair_future)
+            pair_future = RegFuture(self._connection, loop_register)
+            assert params.post_routine is not None
+            params.post_routine(self, q, pair_future)
 
         # TODO use loop context
-        self._build_cmds_loop_body(post_loop, stop=number, loop_register=loop_register)
+        self._build_cmds_loop_body(
+            post_loop, stop=params.number, loop_register=loop_register
+        )
+        self._mem_mgr.remove_active_register(loop_register)
+        self._mem_mgr.remove_active_register(qubit_reg)
+        self._mem_mgr.remove_active_register(bell_state_reg)
 
     def _add_wait_for_ent_info_cmd(
         self, ent_results_array: Array, pair: operand.Register
@@ -721,7 +743,7 @@ class Builder:
 
     def _get_condition_operand(
         self, value: T_CValue
-    ) -> Tuple[List[ICmd], T_OperandUnion]:
+    ) -> Tuple[List[ICmd], T_ProtoOperand]:
         if isinstance(value, Future):
             # Register for checking branching based on condition
             reg = self._mem_mgr.get_inactive_register(activate=True)
@@ -774,7 +796,7 @@ class Builder:
     ) -> Tuple[List[ICmd], List[BranchLabel]]:
         # Exit label
         exit_label = self._label_mgr.new_label(start_with="IF_EXIT")
-        cond_operands: List[T_OperandUnion] = []
+        cond_operands: List[T_ProtoOperand] = []
         if_start: List[ICmd] = []
 
         temp_regs_to_remove: List[operand.Register] = []
@@ -829,8 +851,8 @@ class Builder:
         else:
             loop_register = register
 
-        if self._mem_mgr.is_register_active(loop_register):
-            raise ValueError("Register used for looping should not already be active")
+        # if self._mem_mgr.is_register_active(loop_register):
+        #     raise ValueError("Register used for looping should not already be active")
         return loop_register
 
     def _loop_get_entry_commands(
@@ -1026,7 +1048,7 @@ class Builder:
         self,
         instruction: GenericInstr,
         virtual_qubit_id: int,
-        n: int = 0,
+        n: Union[int, Template] = 0,
         d: int = 0,
         angle: Optional[float] = None,
     ) -> None:
@@ -1041,7 +1063,11 @@ class Builder:
                 )
             return
         if not (isinstance(n, int) and isinstance(d, int) and n >= 0 and d >= 0):
-            raise ValueError(f"{n} * pi / 2 ^ {d} is not a valid angle specification")
+            # We only allow this if n is a Template
+            if not isinstance(n, Template):
+                raise ValueError(
+                    f"{n} * pi / 2 ^ {d} is not a valid angle specification"
+                )
         register = self._get_qubit_register()
         self._build_cmds_set_register_value(register, virtual_qubit_id)
         rot_command = ICmd(
@@ -1081,7 +1107,12 @@ class Builder:
         self._build_cmds_qfree(source)
 
     def _build_cmds_measure(
-        self, qubit_id: int, future: Union[Future, RegFuture], inplace: bool
+        self,
+        qubit_id: int,
+        future: Union[Future, RegFuture],
+        inplace: bool,
+        basis: QubitMeasureBasis = QubitMeasureBasis.Z,
+        rotations: Optional[Tuple[int, int, int]] = None,
     ) -> None:
         if isinstance(self._hardware_config, NVHardwareConfig):
             # If compiling for NV, only virtual ID 0 can be used to measure a qubit.
@@ -1092,10 +1123,32 @@ class Builder:
         outcome_reg = self._mem_mgr.get_new_meas_outcome_register()
         qubit_reg = self._get_qubit_register()
         self._build_cmds_set_register_value(qubit_reg, qubit_id)
-        meas_command = ICmd(
-            instruction=GenericInstr.MEAS,
-            operands=[qubit_reg, outcome_reg],
-        )
+
+        # use denominator 4 since we always treat angles as multiples of pi/(2^4)
+        denom = 4
+
+        if rotations is not None:
+            x1, y, x2 = rotations
+            meas_command = ICmd(
+                instruction=GenericInstr.MEAS_BASIS,
+                operands=[qubit_reg, outcome_reg, x1, y, x2, denom],
+            )
+        elif basis == QubitMeasureBasis.X:
+            meas_command = ICmd(
+                instruction=GenericInstr.MEAS_BASIS,
+                operands=[qubit_reg, outcome_reg, 0, 24, 0, denom],  # -pi/2 Y rotation
+            )
+        elif basis == QubitMeasureBasis.Y:
+            meas_command = ICmd(
+                instruction=GenericInstr.MEAS_BASIS,
+                operands=[qubit_reg, outcome_reg, 8, 0, 0, denom],  # pi/2 X rotation
+            )
+        else:
+            meas_command = ICmd(
+                instruction=GenericInstr.MEAS,
+                operands=[qubit_reg, outcome_reg],
+            )
+
         if not inplace:
             free_commands = [
                 ICmd(
@@ -1243,13 +1296,35 @@ class Builder:
         if isinstance(self._hardware_config, NVHardwareConfig):
             for q in self._mem_mgr.get_active_qubits():
                 # Find a free qubit
-                new_virtual_address = self._mem_mgr.get_new_qubit_address()
                 if q.qubit_id == virtual_address:
-                    # Virtual address is already used. Move it to the new virtual address.
+                    # Virtual address is already used. Move it to a new virtual address.
                     # NOTE: this assumes that the new virtual address is *not* currently used.
-                    self._build_cmds_move_qubit(
-                        source=virtual_address, target=new_virtual_address
-                    )
+                    new_virtual_address = self._mem_mgr.get_new_qubit_address()
+
+                    # Check if the existing qubit was just initialized. If so, we can
+                    # simply change the address used for initialization instead of
+                    # adding move instructions. I.e. instead of
+                    # "init old address" + "move to new address", we just have
+                    # "init new address"
+                    pending_commands = self.subrt_pop_all_pending_commands()
+                    if len(pending_commands) >= 3:
+                        # Check last 3 commands to see if the qubit was just initialized.
+                        # ('type: ignore' since mypy isn't smart enough)
+                        if (
+                            all(isinstance(cmd, ICmd) for cmd in pending_commands[-3:])
+                            and pending_commands[-3].instruction == GenericInstr.SET  # type: ignore
+                            and pending_commands[-2].instruction == GenericInstr.QALLOC  # type: ignore
+                            and pending_commands[-1].instruction == GenericInstr.INIT  # type: ignore
+                        ):
+                            # Update the SET command with the new address.
+                            pending_commands[-3].operands[1] = new_virtual_address  # type: ignore
+                        self.subrt_add_pending_commands(pending_commands)
+                    else:
+                        self.subrt_add_pending_commands(pending_commands)
+                        # Move the existing qubit from the old to the new address.
+                        self._build_cmds_move_qubit(
+                            source=virtual_address, target=new_virtual_address
+                        )
                     # From now on, the original qubit should be referred to with the new virtual address.
                     q.qubit_id = new_virtual_address
 
@@ -1269,6 +1344,85 @@ class Builder:
             loop_register=index_reg,
         )
 
+    def _build_cmds_epr_keep_corrections_single_pair(
+        self, bell_state: RegFuture, qubit_reg: operand.Register
+    ) -> None:
+        with bell_state.if_eq(BellState.PHI_MINUS.value):  # Phi- -> apply Z-gate
+            correction_cmds = [
+                ICmd(instruction=GenericInstr.ROT_Z, operands=[qubit_reg, 16, 4])
+            ]
+            self.subrt_add_pending_commands(correction_cmds)  # type: ignore
+        with bell_state.if_eq(BellState.PSI_PLUS.value):  # Psi+ -> apply X-gate
+            correction_cmds = [
+                ICmd(instruction=GenericInstr.ROT_X, operands=[qubit_reg, 16, 4])
+            ]
+            self.subrt_add_pending_commands(correction_cmds)  # type: ignore
+        with bell_state.if_eq(
+            BellState.PSI_MINUS.value
+        ):  # Psi- -> apply X-gate and Z-gate
+            correction_cmds = [
+                ICmd(instruction=GenericInstr.ROT_X, operands=[qubit_reg, 16, 4]),
+                ICmd(instruction=GenericInstr.ROT_Z, operands=[qubit_reg, 16, 4]),
+            ]
+            self.subrt_add_pending_commands(correction_cmds)  # type: ignore
+
+    def _build_cmds_epr_keep_corrections(
+        self, qubit_ids_array: Array, ent_results_array: Array, params: EntRequestParams
+    ) -> None:
+        qubit_reg = self._mem_mgr.get_inactive_register(activate=True)
+        bell_state_reg = self._mem_mgr.get_inactive_register(activate=True)
+
+        loop_register = self._mem_mgr.get_inactive_register()
+
+        def loop(conn: BaseNetQASMConnection, index: RegFuture):
+            qubit_reg_cmds = qubit_ids_array.get_future_index(index).get_load_commands(
+                qubit_reg
+            )
+            self.subrt_add_pending_commands(qubit_reg_cmds)
+            bell_state = self._get_raw_bell_state(
+                ent_results_array, index, bell_state_reg
+            )
+            self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
+
+        self._build_cmds_loop_body(
+            loop, stop=params.number, loop_register=loop_register
+        )
+
+        self._mem_mgr.remove_active_register(qubit_reg)
+        self._mem_mgr.remove_active_register(bell_state_reg)
+
+    def _get_raw_bell_state(
+        self, results_array: Array, index: RegFuture, target_reg: operand.Register
+    ) -> RegFuture:
+        # calculate index in array
+
+        index_reg = self._mem_mgr.get_inactive_register(activate=True)
+        # set index_reg to index (in single element) of raw_bell_state
+        self.subrt_add_pending_command(
+            ICmd(
+                GenericInstr.SET, operands=[index_reg, SER_RESPONSE_KEEP_IDX_BELL_STATE]
+            )
+        )
+
+        # add index*elt-size to it to arrive in correct element
+        def add_loop(_1, _2):
+            self.subrt_add_pending_command(
+                ICmd(
+                    GenericInstr.ADD,
+                    operands=[index_reg, index_reg, SER_RESPONSE_KEEP_LEN],
+                )
+            )
+
+        self._build_cmds_loop_body(add_loop, stop=index.reg)  # type: ignore
+
+        bell_state_reg_cmds = results_array.get_future_index(
+            index_reg
+        ).get_load_commands(target_reg)
+        self.subrt_add_pending_commands(bell_state_reg_cmds)
+
+        self._mem_mgr.remove_active_register(index_reg)
+        return RegFuture(self._connection, target_reg)
+
     def _build_cmds_epr_create_keep(
         self,
         create_args_array: Array,
@@ -1303,6 +1457,11 @@ class Builder:
 
         self.subrt_add_pending_commands(wait_cmds)  # type: ignore
 
+        if wait_all and params.expect_phi_plus:
+            self._build_cmds_epr_keep_corrections(
+                qubit_ids_array, ent_results_array, params
+            )
+
     def _build_cmds_epr_recv_keep(
         self,
         qubit_ids_array: Array,
@@ -1335,6 +1494,11 @@ class Builder:
 
         self.subrt_add_pending_commands(wait_cmds)  # type: ignore
 
+        if wait_all and params.expect_phi_plus:
+            self._build_cmds_epr_keep_corrections(
+                qubit_ids_array, ent_results_array, params
+            )
+
     def _build_cmds_epr_create_measure(
         self,
         create_args_array: Array,
@@ -1484,7 +1648,12 @@ class Builder:
         loop_register = self._loop_get_register(loop_register)
         pre_commands = self.subrt_pop_all_pending_commands()
 
-        self._mem_mgr.add_active_register(loop_register)
+        loop_register_already_activated: bool = self._mem_mgr.is_register_active(
+            loop_register
+        )
+
+        if not loop_register_already_activated:
+            self._mem_mgr.add_active_register(loop_register)
         # evaluate body (will add pending commands)
         body(
             self._connection,
@@ -1500,7 +1669,8 @@ class Builder:
             step=step,
             loop_register=loop_register,
         )
-        self._mem_mgr.remove_active_register(loop_register)
+        if not loop_register_already_activated:
+            self._mem_mgr.remove_active_register(loop_register)
 
     def _build_cmds_loop(
         self,
@@ -1677,15 +1847,17 @@ class Builder:
             virtual_qubit_ids = [q.qubit_id for q in qubit_futures]
         qubit_ids_array: Array = self.alloc_array(init_values=virtual_qubit_ids)  # type: ignore
 
-        wait_all = True
+        single_comm_qubit: bool = (
+            self._hardware_config is not None
+            and self._hardware_config.comm_qubit_count == 1
+        )
+
         # If there is a post routine, handle pairs one by one.
         # If there is only one comm qubit, handle pairs one by one.
-        if (
-            params.post_routine is not None
-            or self._hardware_config is not None
-            and self._hardware_config.comm_qubit_count == 1
-        ):
+        if params.post_routine is not None or single_comm_qubit:
             wait_all = False
+        else:
+            wait_all = True
 
         if reset_results_array:
             self._build_cmds_undefine_array(ent_results_array)
@@ -1707,23 +1879,18 @@ class Builder:
                 qubit_ids_array, ent_results_array, wait_all, params
             )
 
-        if (
-            not params.sequential
-            and self._hardware_config is not None
-            and self._hardware_config.comm_qubit_count == 1
-        ):
+        if params.post_routine is None and single_comm_qubit:
             self._build_cmds_wait_move_epr_to_mem(
-                number=params.number, ent_results_array=ent_results_array
+                params=params, ent_results_array=ent_results_array
             )
 
         # Construct and add NetQASM instructions for post routine
         if params.post_routine:
             self._build_cmds_post_epr(
                 qubit_ids_array,
-                params.number,
+                params,
                 ent_results_array,
                 EPRType.K,
-                params.post_routine,
             )
 
         return qubit_futures, ent_results_array
@@ -1757,7 +1924,7 @@ class Builder:
         else:
             self._build_cmds_epr_recv_measure(ent_results_array, wait_all, params)
 
-        results = deserialize_epr_measure_results(params.number, ent_results_array)
+        results = deserialize_epr_measure_results(params, ent_results_array, role)
         return results
 
     def sdk_epr_rsp_create(
@@ -1786,7 +1953,9 @@ class Builder:
             create_args_array, ent_results_array, wait_all, params
         )
 
-        return deserialize_epr_measure_results(params.number, ent_results_array)
+        return deserialize_epr_measure_results(
+            params, ent_results_array, role=EPRRole.CREATE
+        )
 
     def sdk_epr_rsp_recv(
         self,
@@ -1823,7 +1992,7 @@ class Builder:
             qubit_ids_array, ent_results_array, wait_all, params
         )
 
-        epr_results = deserialize_epr_keep_results(params.number, ent_results_array)
+        epr_results = deserialize_epr_keep_results(params, ent_results_array)
         return qubit_futures, epr_results
 
     def sdk_create_epr_keep(
@@ -1839,7 +2008,7 @@ class Builder:
                     role=EPRRole.CREATE, params=params
                 )
 
-                results = deserialize_epr_keep_results(params.number, result_array)
+                results = deserialize_epr_keep_results(params, result_array)
 
                 duration = results[-1].generation_duration
                 max_time = NVEprCompiler.get_max_time_for_fidelity(
@@ -1867,7 +2036,7 @@ class Builder:
         else:
             # otherwise, just do the operation once
             qubits, result_array = self.sdk_epr_keep(role=EPRRole.CREATE, params=params)
-            results = deserialize_epr_keep_results(params.number, result_array)
+            results = deserialize_epr_keep_results(params, result_array)
             return qubits, results
 
     def sdk_recv_epr_keep(
@@ -1884,7 +2053,7 @@ class Builder:
                     role=EPRRole.RECV, params=params, reset_results_array=True
                 )
 
-                results = deserialize_epr_keep_results(params.number, result_array)
+                results = deserialize_epr_keep_results(params, result_array)
 
                 duration = results[-1].generation_duration
                 max_time = NVEprCompiler.get_max_time_for_fidelity(
@@ -1907,7 +2076,7 @@ class Builder:
         else:
             # otherwise, just do the operation once
             qubits, result_array = self.sdk_epr_keep(role=EPRRole.RECV, params=params)
-            results = deserialize_epr_keep_results(params.number, result_array)
+            results = deserialize_epr_keep_results(params, result_array)
             return qubits, results
 
     def sdk_create_epr_measure(
